//////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2007 Xilinx, Inc.
// This design is confidential and proprietary of Xilinx, All Rights Reserved.
//////////////////////////////////////////////////////////////////////////////
//   ____  ____
//  /   /\/   /
// /___/  \  /   Vendor:        Xilinx
// \   \   \/    Version:       1.0.0
//  \   \        Filename:      fbctime.v
//  /   /        Date Created:  July 1, 2007
// /___/   /\    Last Modified: July 1, 2007
// \   \  /  \
//  \___\/\___\
//
// Devices:   Spartan-3 Generation FPGA
// Purpose:   Programmable video timing generation circuit
// Contact:   crabill@xilinx.com
// Reference: None
//
// Revision History:
//   Rev 1.0.0 - (crabill) First created July 1, 2007.
//
//////////////////////////////////////////////////////////////////////////////
//
// LIMITED WARRANTY AND DISCLAIMER. These designs are provided to you "as is".
// Xilinx and its licensors make and you receive no warranties or conditions,
// express, implied, statutory or otherwise, and Xilinx specifically disclaims
// any implied warranties of merchantability, non-infringement, or fitness for
// a particular purpose. Xilinx does not warrant that the functions contained
// in these designs will meet your requirements, or that the operation of
// these designs will be uninterrupted or error free, or that defects in the
// designs will be corrected. Furthermore, Xilinx does not warrant or make any
// representations regarding use or the results of the use of the designs in
// terms of correctness, accuracy, reliability, or otherwise.
//
// LIMITATION OF LIABILITY. In no event will Xilinx or its licensors be liable
// for any loss of data, lost profits, cost or procurement of substitute goods
// or services, or for any special, incidental, consequential, or indirect
// damages arising from the use or operation of the designs or accompanying
// documentation, however caused and on any theory of liability. This
// limitation will apply even if Xilinx has been advised of the possibility
// of such damage. This limitation shall apply not-withstanding the failure
// of the essential purpose of any limited remedies herein.
//
//////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2007 Xilinx, Inc.
// This design is confidential and proprietary of Xilinx, All Rights Reserved.
//////////////////////////////////////////////////////////////////////////////

`timescale 1 ns / 1 ps

module fbctime
  (
  input  wire [11:0] tc_hsblnk,
  input  wire [11:0] tc_hssync,
  input  wire [11:0] tc_hesync,
  input  wire [11:0] tc_heblnk,

  output reg  [11:0] hcount,
  output reg         hsync,
  output reg         hblnk,

  input  wire [11:0] tc_vsblnk,
  input  wire [11:0] tc_vssync,
  input  wire [11:0] tc_vesync,
  input  wire [11:0] tc_veblnk,

  output reg  [11:0] vcount,
  output reg         vsync,
  output reg         vblnk,

  output reg  [11:0] vfetch,
  output reg         active,

  input  wire        clk
  );

  //******************************************************************//
  // This logic describes a 12-bit horizontal position counter.       //
  //******************************************************************//
 
  reg    [11:0] hpos_end = 0;
  reg    [11:0] hpos_cnt = 0;
  reg           hpos_clr = 0;
  reg           hpos_sta = 1;

  always @(posedge clk)
  begin : hcounter
    hpos_end <= tc_heblnk + 12'b1111_1111_1111;
    if (hpos_clr) hpos_cnt <= 12'b0000_0000_0000;
    else hpos_cnt <= hpos_cnt + 12'b0000_0000_0001;
    hpos_clr <= (hpos_cnt == hpos_end);
    hpos_sta <= hpos_clr;
  end

  //******************************************************************//
  // This logic describes a 12-bit vertical position counter.         //
  //******************************************************************//

  reg    [11:0] vpos_end = 0;
  reg    [11:0] vpos_cnt = 0;
  wire          vpos_clr;

  reg    [11:0] vfet_cnt = 0;
  reg           vpos_lst = 0;
  reg           vpos_pre = 0;
  wire          vadv_clr;
  reg           go_fetch = 1;
  reg           vfet_sob = 0;
  wire          vfet_clr;

  always @(posedge clk)
  begin : vcounter
    vpos_end <= tc_veblnk + 12'b1111_1111_1111;
    if (vpos_clr) vpos_cnt <= 12'b0000_0000_0000;
    else if (hpos_clr) vpos_cnt <= vpos_cnt + 12'b0000_0000_0001;
    if (vadv_clr) vfet_cnt <= 12'b0000_0000_0000;
    else if (hpos_clr) vfet_cnt <= vfet_cnt + 12'b0000_0000_0001;
    if (vadv_clr) go_fetch <= 1'b1;
    else if (vfet_clr) go_fetch <= 1'b0;
    // this is late one clock, but
    // is not used at those times
    // since gated with hpos_clr
    vpos_lst <= (vpos_cnt == tc_veblnk);
    vpos_pre <= (vpos_cnt == vpos_end);
    vfet_sob <= (vfet_cnt == tc_vsblnk);
  end

  assign vpos_clr = vpos_lst && hpos_clr;
  assign vadv_clr = vpos_pre && hpos_clr;
  assign vfet_clr = vfet_sob && hpos_clr;

  //******************************************************************//
  // This is the logic for the horizontal outputs.  Active video is   //
  // always started when the horizontal count is zero.  Example:      //
  //                                                                  //
  // tc_hsblnk = 03                                                   //
  // tc_hssync = 07                                                   //
  // tc_hesync = 11                                                   //
  // tc_heblnk = 15 (htotal)                                          //
  //                                                                  //
  // hcount   00 01 02 03 04 05 06 07 08 09 10 11 12 13 14 15         //
  // hsync    ________________________------------____________        //
  // hblnk    ____________------------------------------------        //
  //                                                                  //
  // hsync time  = (tc_hesync - tc_hssync) pixels                     //
  // hblnk time  = (tc_heblnk - tc_hsblnk) pixels                     //
  // active time = (tc_hsblnk + 1) pixels                             //
  //                                                                  //
  //******************************************************************//

  always @(posedge clk) hcount <= hpos_cnt;
  always @(posedge clk) hblnk <= (hpos_cnt > tc_hsblnk);
  always @(posedge clk) hsync <= (hpos_cnt > tc_hssync) && (hpos_cnt <= tc_hesync);

  //******************************************************************//
  // This is the logic for the vertical outputs.  Active video is     //
  // always started when the vertical count is zero.  Example:        //
  //                                                                  //
  // tc_vsblnk = 03                                                   //
  // tc_vssync = 07                                                   //
  // tc_vesync = 11                                                   //
  // tc_veblnk = 15 (vtotal)                                          //
  //                                                                  //
  // vcount   00 01 02 03 04 05 06 07 08 09 10 11 12 13 14 15         //
  // vsync    ________________________------------____________        //
  // vblnk    ____________------------------------------------        //
  //                                                                  //
  // vsync time  = (tc_vesync - tc_vssync) lines                      //
  // vblnk time  = (tc_veblnk - tc_vsblnk) lines                      //
  // active time = (tc_vsblnk + 1) lines                              //
  //                                                                  //
  //******************************************************************//
 
  always @(posedge clk) vcount <= vpos_cnt;
  always @(posedge clk) vblnk <= (vpos_cnt > tc_vsblnk);
  always @(posedge clk) vsync <= (vpos_cnt > tc_vssync) && (vpos_cnt <= tc_vesync);

  //******************************************************************//
  //                                                                  //
  //******************************************************************//

  always @(posedge clk) active <= hpos_sta & go_fetch;
  always @(posedge clk) vfetch <= vfet_cnt;

  //******************************************************************//
  //                                                                  //
  //******************************************************************//

endmodule