//////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2007 Xilinx, Inc.
// This design is confidential and proprietary of Xilinx, All Rights Reserved.
//////////////////////////////////////////////////////////////////////////////
//   ____  ____
//  /   /\/   /
// /___/  \  /   Vendor:        Xilinx
// \   \   \/    Version:       1.0.0
//  \   \        Filename:      rs232reg.v
//  /   /        Date Created:  July 1, 2007
// /___/   /\    Last Modified: July 1, 2007
// \   \  /  \
//  \___\/\___\
//
// Devices:   Spartan-3 Generation FPGA
// Purpose:   Regulator control via RS232
// Contact:   crabill@xilinx.com
// Reference: None
//
// Revision History:
//   Rev 1.0.0 - (crabill) Created July 1, 2007 for PCB revision D.
//
//////////////////////////////////////////////////////////////////////////////
//
// LIMITED WARRANTY AND DISCLAIMER. These designs are provided to you "as is".
// Xilinx and its licensors make and you receive no warranties or conditions,
// express, implied, statutory or otherwise, and Xilinx specifically disclaims
// any implied warranties of merchantability, non-infringement, or fitness for
// a particular purpose. Xilinx does not warrant that the functions contained
// in these designs will meet your requirements, or that the operation of
// these designs will be uninterrupted or error free, or that defects in the
// designs will be corrected. Furthermore, Xilinx does not warrant or make any
// representations regarding use or the results of the use of the designs in
// terms of correctness, accuracy, reliability, or otherwise.
//
// LIMITATION OF LIABILITY. In no event will Xilinx or its licensors be liable
// for any loss of data, lost profits, cost or procurement of substitute goods
// or services, or for any special, incidental, consequential, or indirect
// damages arising from the use or operation of the designs or accompanying
// documentation, however caused and on any theory of liability. This
// limitation will apply even if Xilinx has been advised of the possibility
// of such damage. This limitation shall apply not-withstanding the failure
// of the essential purpose of any limited remedies herein.
//
//////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2007 Xilinx, Inc.
// This design is confidential and proprietary of Xilinx, All Rights Reserved.
//////////////////////////////////////////////////////////////////////////////

`timescale 1 ns / 1 ps

module rs232reg
  (
  input  wire        clk,
  input  wire        reset,
 
  input  wire        rs232_dce_rxd,
  output wire        rs232_dce_txd,

  inout  wire        reg1_sda,
  inout  wire        reg1_scl,
  inout  wire        reg2_sda,
  inout  wire        reg2_scl
  );

  //******************************************************************//
  // Instantiate PicoBlaze and the Instruction ROM.                   // 
  //******************************************************************//

  wire    [9:0] address;
  wire   [17:0] instruction;
  wire    [7:0] port_id;
  wire    [7:0] out_port;
  reg     [7:0] in_port;
  wire          write_strobe;
  wire          read_strobe;

  kcpsm3 kcpsm3_inst (
    .address(address),
    .instruction(instruction),
    .port_id(port_id),
    .write_strobe(write_strobe),
    .out_port(out_port),
    .read_strobe(read_strobe),
    .in_port(in_port),
    .interrupt(1'b0),
    .interrupt_ack(),
    .reset(reset),
    .clk(clk));

  vregsw vregsw_inst (
    .address(address),
    .instruction(instruction),
    .clk(clk));

  //******************************************************************//
  // Generate bit time enable for serial ports to run at 115200 bps.  //
  //******************************************************************//

  reg     [8:0] en_16_x_cntr;
  wire          en_16_x_baud;

  assign en_16_x_baud = (en_16_x_cntr == 26);

  always @(posedge clk)
  begin
    if (en_16_x_baud) en_16_x_cntr <= 0;
    else en_16_x_cntr <= en_16_x_cntr + 1;
  end

  //******************************************************************//
  // Implement DCE serial port (used for main communication).         //
  // Data Port is 00h for read and write of data.                     //
  // Status Port is 01h to sense tx buffer full.                      //
  // Status Port is 02h to sense rx data present.                     //
  //******************************************************************//

  reg           dce_rd;
  wire          dce_wr;
  wire          dce_tx_full;
  wire          dce_rx_pres;
  wire    [7:0] dce_rx_data;

  assign dce_wr = write_strobe & (port_id == 8'h00);
  always @(posedge clk) dce_rd <= read_strobe & (port_id == 8'h00);

  uart_tx uart_tx_dce (
    .data_in(out_port), 
    .write_buffer(dce_wr),
    .reset_buffer(reset),
    .en_16_x_baud(en_16_x_baud),
    .serial_out(rs232_dce_txd),
    .buffer_full(dce_tx_full),
    .buffer_half_full(),
    .clk(clk)
    );

  uart_rx uart_rx_dce (
    .serial_in(rs232_dce_rxd),
    .data_out(dce_rx_data),
    .read_buffer(dce_rd),
    .reset_buffer(reset),
    .en_16_x_baud(en_16_x_baud),
    .buffer_data_present(dce_rx_pres),
    .buffer_full(),
    .buffer_half_full(),
    .clk(clk)
    );  

  //******************************************************************//
  // Implement peripheral subsystem (mostly bit-bang serial).         //
  // Port 18h is read-write one-hot i2c device select:                //
  //          00 -- nothing selected                                  //
  //          01 -- regulator one selected                            //
  //          02 -- regulator two selected                            //
  // Port 19h is read-write serial bit-bang register to communicate   //
  // with the i2c-enabled power regulators.  Be careful using this.   //
  //          bit 0 -- serial clock (SCL)                             //
  //          bit 1 -- serial data (SDA)                              //
  //******************************************************************//

  wire          i2c_select_wr;
  wire          i2c_sdascl_wr;
  reg     [1:0] i2c_select;
  reg     [1:0] i2c_sdascl;
  reg     [1:0] i2c_sense;

  reg           reg1_sda_drivelo;
  reg           reg1_scl_drivelo;
  reg           reg2_sda_drivelo;
  reg           reg2_scl_drivelo;

  assign i2c_select_wr = write_strobe & (port_id == 8'h18);
  assign i2c_sdascl_wr = write_strobe & (port_id == 8'h19);

  always @(posedge clk)
  begin
    if (reset)
    begin
      i2c_select <= 2'b00;
      i2c_sdascl <= 2'b11;
    end
    else
    begin
      if (i2c_select_wr) i2c_select <= out_port[1:0];
      if (i2c_sdascl_wr) i2c_sdascl <= out_port[1:0];
    end
  end

  always @(posedge clk)
  begin
    if (reset)
    begin
      reg1_sda_drivelo <= 1'b0;
      reg1_scl_drivelo <= 1'b0;
      reg2_sda_drivelo <= 1'b0;
      reg2_scl_drivelo <= 1'b0;
    end
    else
    begin
      reg1_sda_drivelo <= (i2c_select[0] && !i2c_sdascl[1]);
      reg1_scl_drivelo <= (i2c_select[0] && !i2c_sdascl[0]);
      reg2_sda_drivelo <= (i2c_select[1] && !i2c_sdascl[1]);
      reg2_scl_drivelo <= (i2c_select[1] && !i2c_sdascl[0]);
    end
  end

  assign reg1_sda = reg1_sda_drivelo ? 1'b0 : 1'bz;
  assign reg1_scl = reg1_scl_drivelo ? 1'b0 : 1'bz;
  assign reg2_sda = reg2_sda_drivelo ? 1'b0 : 1'bz;
  assign reg2_scl = reg2_scl_drivelo ? 1'b0 : 1'bz;

  always @*
  begin
    case (i2c_select)
      2'b01: i2c_sense = {reg1_sda, reg1_scl};
      2'b10: i2c_sense = {reg2_sda, reg2_scl};
      default: i2c_sense = 2'b11;
    endcase
  end

  //******************************************************************//
  // Implement input port logic.                                      //
  //******************************************************************//

  always @(posedge clk)
  begin
    case (port_id)
      8'h00: in_port <= dce_rx_data;
      8'h01: in_port <= {7'b0,dce_tx_full};
      8'h02: in_port <= {7'b0,dce_rx_pres};
      8'h18: in_port <= {6'b000000, i2c_select};
      8'h19: in_port <= {6'b000000, i2c_sense};
      default: in_port <= 8'h00;
    endcase
  end

  //******************************************************************//
  //                                                                  //
  //******************************************************************//

endmodule